@@ -46,6 +46,8 @@
 #endif
 #endif
 
+#include <netinet/ip.h>
+
 int
 lws_send_pipe_choked(struct lws *wsi)
 {
@@ -160,7 +162,7 @@ lws_plat_set_socket_options(struct lws_vhost *vhost, int fd, int unix_skt)
 	if (!unix_skt && vhost->bind_iface && vhost->iface) {
 		lwsl_info("binding listen skt to %s using SO_BINDTODEVICE\n", vhost->iface);
 		if (setsockopt(fd, SOL_SOCKET, SO_BINDTODEVICE, vhost->iface,
-				strlen(vhost->iface)) < 0) {
+				(socklen_t)strlen(vhost->iface)) < 0) {
 			lwsl_warn("Failed to bind to device %s\n", vhost->iface);
 			return 1;
 		}
@@ -188,6 +190,80 @@ lws_plat_set_socket_options(struct lws_vhost *vhost, int fd, int unix_skt)
 	return lws_plat_set_nonblocking(fd);
 }
 
+static const int ip_opt_lws_flags[] = {
+	LCCSCF_IP_LOW_LATENCY, LCCSCF_IP_HIGH_THROUGHPUT,
+	LCCSCF_IP_HIGH_RELIABILITY
+#if !defined(__OpenBSD__)
+	, LCCSCF_IP_LOW_COST
+#endif
+}, ip_opt_val[] = {
+	IPTOS_LOWDELAY, IPTOS_THROUGHPUT, IPTOS_RELIABILITY
+#if !defined(__OpenBSD__) && !defined(__sun)
+	, IPTOS_MINCOST
+#endif
+};
+#if !defined(LWS_WITH_NO_LOGS)
+static const char *ip_opt_names[] = {
+	"LOWDELAY", "THROUGHPUT", "RELIABILITY"
+#if !defined(__OpenBSD__) && !defined(__sun)
+	, "MINCOST"
+#endif
+};
+#endif
+
+int
+lws_plat_set_socket_options_ip(lws_sockfd_type fd, uint8_t pri, int lws_flags)
+{
+	int optval = (int)pri, ret = 0, n;
+	socklen_t optlen = sizeof(optval);
+#if !defined(LWS_WITH_NO_LOGS)
+	int en;
+#endif
+
+#if !defined(__APPLE__) && \
+      !defined(__FreeBSD__) && !defined(__FreeBSD_kernel__) &&        \
+      !defined(__NetBSD__) && \
+      !defined(__OpenBSD__) && \
+      !defined(__sun) && \
+      !defined(__HAIKU__) && \
+      !defined(__CYGWIN__)
+
+	/* the BSDs don't have SO_PRIORITY */
+
+	if (pri) { /* 0 is the default already */
+		if (setsockopt(fd, SOL_SOCKET, SO_PRIORITY,
+				(const void *)&optval, optlen) < 0) {
+#if !defined(LWS_WITH_NO_LOGS)
+			en = errno;
+			lwsl_warn("%s: unable to set socket pri %d: errno %d\n",
+				  __func__, (int)pri, en);
+#endif
+			ret = 1;
+		} else
+			lwsl_notice("%s: set pri %u\n", __func__, pri);
+	}
+#endif
+
+	for (n = 0; n < 4; n++) {
+		if (!(lws_flags & ip_opt_lws_flags[n]))
+			continue;
+
+		optval = (int)ip_opt_val[n];
+		if (setsockopt(fd, IPPROTO_IP, IP_TOS, (const void *)&optval,
+			       optlen) < 0) {
+#if !defined(LWS_WITH_NO_LOGS)
+			en = errno;
+			lwsl_warn("%s: unable to set %s: errno %d\n", __func__,
+				  ip_opt_names[n], en);
+#endif
+			ret = 1;
+		} else
+			lwsl_notice("%s: set ip flag %s\n", __func__,
+				    ip_opt_names[n]);
+	}
+
+	return ret;
+}
 
 /* cast a struct sockaddr_in6 * into addr for ipv6 */
 
@@ -260,23 +336,16 @@ lws_interface_to_sa(int ipv6, const char *ifname, struct sockaddr_in *addr,
 
 	freeifaddrs(ifr);
 
-	if (rc) {
-		/* check if bind to IP address */
-#ifdef LWS_WITH_IPV6
-		if (inet_pton(AF_INET6, ifname, &addr6->sin6_addr) == 1)
-			rc = LWS_ITOSA_USABLE;
-		else
-#endif
-		if (inet_pton(AF_INET, ifname, &addr->sin_addr) == 1)
-			rc = LWS_ITOSA_USABLE;
-	}
+	if (rc &&
+	    !lws_sa46_parse_numeric_address(ifname, (lws_sockaddr46 *)addr))
+		rc = LWS_ITOSA_USABLE;
 
 	return rc;
 }
 
 
 const char *
-lws_plat_inet_ntop(int af, const void *src, char *dst, int cnt)
+lws_plat_inet_ntop(int af, const void *src, char *dst, socklen_t cnt)
 {
 	return inet_ntop(af, src, dst, cnt);
 }
@@ -310,8 +379,8 @@ lws_plat_ifname_to_hwaddr(int fd, const char *ifname, uint8_t *hwaddr, int len)
 }
 
 int
-lws_plat_rawudp_broadcast(uint8_t *p, const uint8_t *canned, int canned_len,
-			  int n, int fd, const char *iface)
+lws_plat_rawudp_broadcast(uint8_t *p, const uint8_t *canned, size_t canned_len,
+			  size_t n, int fd, const char *iface)
 {
 #if defined(__linux__)
 	struct sockaddr_ll sll;
@@ -320,8 +389,8 @@ lws_plat_rawudp_broadcast(uint8_t *p, const uint8_t *canned, int canned_len,
 
 	memcpy(p, canned, canned_len);
 
-	p[2] = n >> 8;
-	p[3] = n;
+	p[2] = (uint8_t)(n >> 8);
+	p[3] = (uint8_t)(n);
 
 	while (p16 < (uint16_t *)(p + 20))
 		ucs += ntohs(*p16++);
@@ -329,19 +398,19 @@ lws_plat_rawudp_broadcast(uint8_t *p, const uint8_t *canned, int canned_len,
 	ucs += ucs >> 16;
 	ucs ^= 0xffff;
 
-	p[10] = ucs >> 8;
-	p[11] = ucs;
-	p[24] = (n - 20) >> 8;
-	p[25] = (n - 20);
+	p[10] = (uint8_t)(ucs >> 8);
+	p[11] = (uint8_t)(ucs);
+	p[24] = (uint8_t)((n - 20) >> 8);
+	p[25] = (uint8_t)((n - 20));
 
 	memset(&sll, 0, sizeof(sll));
 	sll.sll_family = AF_PACKET;
 	sll.sll_protocol = htons(0x800);
 	sll.sll_halen = 6;
-	sll.sll_ifindex = if_nametoindex(iface);
+	sll.sll_ifindex = (int)if_nametoindex(iface);
 	memset(sll.sll_addr, 0xff, 6);
 
-	return sendto(fd, p, n, 0, (struct sockaddr *)&sll, sizeof(sll));
+	return (int)sendto(fd, p, n, 0, (struct sockaddr *)&sll, sizeof(sll));
 #else
 	lwsl_err("%s: UNIMPLEMENTED on this platform\n", __func__);
 
@@ -405,56 +474,60 @@ lws_plat_BINDTODEVICE(lws_sockfd_type fd, const char *ifname)
 }
 
 int
-lws_plat_ifconfig_ip(const char *ifname, int fd, uint8_t *ip, uint8_t *mask_ip,
-			uint8_t *gateway_ip)
+lws_plat_ifconfig(int fd, lws_dhcpc_ifstate_t *is)
 {
 #if defined(__linux__)
-	struct sockaddr_in sin;
 	struct rtentry route;
 	struct ifreq ifr;
 
 	memset(&ifr, 0, sizeof(ifr));
 	memset(&route, 0, sizeof(route));
-	memset(&sin, 0, sizeof(sin));
 
-	lws_strncpy(ifr.ifr_name, ifname, IFNAMSIZ);
+	lws_strncpy(ifr.ifr_name, is->ifname, IFNAMSIZ);
 
-	lws_plat_if_up(ifname, fd, 0);
+	lws_plat_if_up(is->ifname, fd, 0);
 
-	sin.sin_family = AF_INET;
-	sin.sin_addr.s_addr = htonl(*(uint32_t *)ip);
-
-	memcpy(&ifr.ifr_addr, &sin, sizeof(struct sockaddr));
+	memcpy(&ifr.ifr_addr, &is->sa46[LWSDH_SA46_IP], sizeof(struct sockaddr));
 	if (ioctl(fd, SIOCSIFADDR, &ifr) < 0) {
 		lwsl_err("%s: SIOCSIFADDR fail\n", __func__);
 		return 1;
 	}
 
-	sin.sin_addr.s_addr = htonl(*(uint32_t *)mask_ip);
-	memcpy(&ifr.ifr_addr, &sin, sizeof(struct sockaddr));
-	if (ioctl(fd, SIOCSIFNETMASK, &ifr) < 0) {
-		lwsl_err("%s: SIOCSIFNETMASK fail\n", __func__);
-		return 1;
-	}
+	if (is->sa46[LWSDH_SA46_IP].sa4.sin_family == AF_INET) {
+		struct sockaddr_in sin;
 
-	lws_plat_if_up(ifname, fd, 1);
+		memset(&sin, 0, sizeof(sin));
+		sin.sin_family = AF_INET;
+		sin.sin_addr.s_addr = *(uint32_t *)&is->nums[LWSDH_IPV4_SUBNET_MASK];
+		memcpy(&ifr.ifr_addr, &sin, sizeof(struct sockaddr));
+		if (ioctl(fd, SIOCSIFNETMASK, &ifr) < 0) {
+			lwsl_err("%s: SIOCSIFNETMASK fail\n", __func__);
+			return 1;
+		}
 
-	sin.sin_addr.s_addr = htonl(*(uint32_t *)gateway_ip);
-	memcpy(&route.rt_gateway, &sin, sizeof(struct sockaddr));
+		lws_plat_if_up(is->ifname, fd, 1);
 
-	sin.sin_addr.s_addr = 0;
-	memcpy(&route.rt_dst, &sin, sizeof(struct sockaddr));
-	memcpy(&route.rt_genmask, &sin, sizeof(struct sockaddr));
+		memcpy(&route.rt_gateway,
+		       &is->sa46[LWSDH_SA46_IPV4_ROUTER].sa4,
+		       sizeof(struct sockaddr));
 
-	route.rt_flags = RTF_UP | RTF_GATEWAY;
-	route.rt_metric = 100;
-	route.rt_dev = (char *)ifname;
+		sin.sin_addr.s_addr = 0;
+		memcpy(&route.rt_dst, &sin, sizeof(struct sockaddr));
+		memcpy(&route.rt_genmask, &sin, sizeof(struct sockaddr));
 
-	if (ioctl(fd, SIOCADDRT, &route) < 0) {
-		lwsl_err("%s: SIOCADDRT 0x%x fail: %d\n", __func__,
-			(unsigned int)htonl(*(uint32_t *)gateway_ip), LWS_ERRNO);
-		return 1;
-	}
+		route.rt_flags = RTF_UP | RTF_GATEWAY;
+		route.rt_metric = 100;
+		route.rt_dev = (char *)is->ifname;
+
+		if (ioctl(fd, SIOCADDRT, &route) < 0) {
+			lwsl_err("%s: SIOCADDRT 0x%x fail: %d\n", __func__,
+				(unsigned int)htonl(*(uint32_t *)&is->
+					sa46[LWSDH_SA46_IPV4_ROUTER].
+						sa4.sin_addr.s_addr), LWS_ERRNO);
+			return 1;
+		}
+	} else
+		lws_plat_if_up(is->ifname, fd, 1);
 
 	return 0;
 #else
@@ -464,6 +537,12 @@ lws_plat_ifconfig_ip(const char *ifname, int fd, uint8_t *ip, uint8_t *mask_ip,
 #endif
 }
 
+int
+lws_plat_vhost_tls_client_ctx_init(struct lws_vhost *vhost)
+{
+	return 0;
+}
+
 #if defined(LWS_WITH_MBEDTLS)
 int
 lws_plat_mbedtls_net_send(void *ctx, const uint8_t *buf, size_t len)
@@ -474,7 +553,7 @@ lws_plat_mbedtls_net_send(void *ctx, const uint8_t *buf, size_t len)
 	if (fd < 0)
 		return MBEDTLS_ERR_NET_INVALID_CONTEXT;
 
-	ret = write(fd, buf, len);
+	ret = (int)write(fd, buf, len);
 	if (ret >= 0)
 		return ret;
 
