@@ -1,7 +1,7 @@
 /*
  * lws-api-test-jose - RFC7516 jwe tests
  *
- * Written in 2010-2018 by Andy Green <andy@warmcat.com>
+ * Written in 2010-2020 by Andy Green <andy@warmcat.com>
  *
  * This file is made available under the Creative Commons CC0 1.0
  * Universal Public Domain Dedication.
@@ -134,9 +134,9 @@ test_jwe_a1(struct lws_context *context)
 	/* we require a JOSE-formatted header to do the encryption */
 
 	jwe.jws.map.buf[LJWS_JOSE] = temp;
-	jwe.jws.map.len[LJWS_JOSE] = lws_snprintf(temp, temp_len,
+	jwe.jws.map.len[LJWS_JOSE] = (uint32_t)lws_snprintf(temp, (unsigned int)temp_len,
 			"{\"alg\":\"%s\",\"enc\":\"%s\"}", "RSA-OAEP", "A256GCM");
-	temp_len -= jwe.jws.map.len[LJWS_JOSE];
+	temp_len -= (int)jwe.jws.map.len[LJWS_JOSE];
 
 	/*
 	 * dup the plaintext into the ciphertext element, it will be
@@ -155,7 +155,7 @@ test_jwe_a1(struct lws_context *context)
 	n = lws_gencrypto_bits_to_bytes(jwe.jose.enc_alg->keybits_fixed);
 	if (lws_jws_randomize_element(context, &jwe.jws.map, LJWE_EKEY,
 				      lws_concat_temp(temp, temp_len),
-				      &temp_len, n,
+				      &temp_len, (unsigned int)n,
 				      LWS_JWE_LIMIT_KEY_ELEMENT_BYTES)) {
 		lwsl_err("Problem getting random\n");
 		goto bail;
@@ -500,7 +500,7 @@ test_jwe_ra_ptext_1024(struct lws_context *context, char *jwk_txt, int jwk_len)
 
 	/* reuse the rsa private key from the JWE Appendix 2 test above */
 
-	if (lws_jwk_import(&jwe.jwk, NULL, NULL, jwk_txt, jwk_len) < 0) {
+	if (lws_jwk_import(&jwe.jwk, NULL, NULL, jwk_txt, (unsigned int)jwk_len) < 0) {
 		lwsl_notice("%s: Failed to decode JWK test key\n", __func__);
 		goto bail;
 	}
@@ -530,7 +530,7 @@ test_jwe_ra_ptext_1024(struct lws_context *context, char *jwk_txt, int jwk_len)
 	jwe.jws.map.len[LJWE_JOSE] = (uint32_t)strlen(rsa256a128_jose);
 
 	n = lws_jwe_parse_jose(&jwe.jose, jwe.jws.map.buf[LJWE_JOSE],
-			       jwe.jws.map.len[LJWE_JOSE],
+			       (int)jwe.jws.map.len[LJWE_JOSE],
 			       lws_concat_temp(temp, temp_len), &temp_len);
 	if (n < 0) {
 		lwsl_err("%s: JOSE parse failed\n", __func__);
@@ -559,7 +559,7 @@ test_jwe_ra_ptext_1024(struct lws_context *context, char *jwk_txt, int jwk_len)
 
 	/* now we created the encrypted version, see if we can decrypt it */
 
-	if (lws_jwk_import(&jwe.jwk, NULL, NULL, jwk_txt, jwk_len) < 0) {
+	if (lws_jwk_import(&jwe.jwk, NULL, NULL, jwk_txt, (unsigned int)jwk_len) < 0) {
 		lwsl_notice("%s: Failed to decode JWK test key\n", __func__);
 		goto bail;
 	}
@@ -627,7 +627,7 @@ test_jwe_r256a192_ptext(struct lws_context *context, char *jwk_txt, int jwk_len)
 
 	/* reuse the rsa private key from the JWE Appendix 2 test above */
 
-	if (lws_jwk_import(&jwe.jwk, NULL, NULL, jwk_txt, jwk_len) < 0) {
+	if (lws_jwk_import(&jwe.jwk, NULL, NULL, jwk_txt, (unsigned int)jwk_len) < 0) {
 		lwsl_notice("%s: Failed to decode JWK test key\n", __func__);
 		goto bail;
 	}
@@ -660,7 +660,7 @@ test_jwe_r256a192_ptext(struct lws_context *context, char *jwk_txt, int jwk_len)
 	jwe.jws.map.len[LJWE_JOSE] = (uint32_t)strlen(rsa256a192_jose);
 
 	n = lws_jwe_parse_jose(&jwe.jose, jwe.jws.map.buf[LJWE_JOSE],
-			       jwe.jws.map.len[LJWE_JOSE],
+			       (int)jwe.jws.map.len[LJWE_JOSE],
 			       lws_concat_temp(temp, temp_len), &temp_len);
 	if (n < 0) {
 		lwsl_err("%s: JOSE parse failed\n", __func__);
@@ -688,7 +688,7 @@ test_jwe_r256a192_ptext(struct lws_context *context, char *jwk_txt, int jwk_len)
 	lws_jwe_destroy(&jwe);
 	lws_jwe_init(&jwe, context);
 
-	if (lws_jwk_import(&jwe.jwk, NULL, NULL, jwk_txt, jwk_len) < 0) {
+	if (lws_jwk_import(&jwe.jwk, NULL, NULL, jwk_txt, (unsigned int)jwk_len) < 0) {
 		lwsl_notice("%s: Failed to decode JWK test key\n", __func__);
 		goto bail;
 	}
@@ -759,7 +759,7 @@ test_jwe_r256a256_ptext(struct lws_context *context, char *jwk_txt, int jwk_len)
 
 	/* reuse the rsa private key from the JWE Appendix 2 test above */
 
-	if (lws_jwk_import(&jwe.jwk, NULL, NULL, jwk_txt, jwk_len) < 0) {
+	if (lws_jwk_import(&jwe.jwk, NULL, NULL, jwk_txt, (unsigned int)jwk_len) < 0) {
 		lwsl_notice("%s: Failed to decode JWK test key\n", __func__);
 		goto bail;
 	}
@@ -789,7 +789,7 @@ test_jwe_r256a256_ptext(struct lws_context *context, char *jwk_txt, int jwk_len)
 	}
 
 	jwe.jws.map.buf[LJWE_JOSE] = rsa256a256_jose;
-	jwe.jws.map.len[LJWE_JOSE] = (int)strlen(rsa256a256_jose);
+	jwe.jws.map.len[LJWE_JOSE] = (uint32_t)strlen(rsa256a256_jose);
 
 	n = lws_jwe_parse_jose(&jwe.jose, rsa256a256_jose,
 			       (int)strlen(rsa256a256_jose),
@@ -820,7 +820,7 @@ test_jwe_r256a256_ptext(struct lws_context *context, char *jwk_txt, int jwk_len)
 	lws_jwe_destroy(&jwe);
 	lws_jwe_init(&jwe, context);
 
-	if (lws_jwk_import(&jwe.jwk, NULL, NULL, jwk_txt, jwk_len) < 0) {
+	if (lws_jwk_import(&jwe.jwk, NULL, NULL, jwk_txt, (unsigned int)jwk_len) < 0) {
 		lwsl_notice("%s: Failed to decode JWK test key\n", __func__);
 		goto bail;
 	}
@@ -1993,7 +1993,7 @@ test_akw_decrypt(struct lws_context *context, const char *test_name,
 	/* allowing for trailing padding, confirm the plaintext */
 	if (jwe.jws.map.len[LJWE_CTXT] < strlen(akw_ptext) ||
 	    lws_timingsafe_bcmp(jwe.jws.map.buf[LJWE_CTXT], akw_ptext,
-			        (int)strlen(akw_ptext))) {
+			        (uint32_t)strlen(akw_ptext))) {
 		lwsl_err("%s: plaintext AES decrypt wrong\n", __func__);
 		lwsl_hexdump_notice(akw_ptext, strlen(akw_ptext));
 		lwsl_hexdump_notice(jwe.jws.map.buf[LJWE_CTXT],
@@ -2042,9 +2042,9 @@ test_akw_encrypt(struct lws_context *context, const char *test_name,
 	/* we require a JOSE-formatted header to do the encryption */
 
 	jwe.jws.map.buf[LJWS_JOSE] = temp;
-	jwe.jws.map.len[LJWS_JOSE] = lws_snprintf(temp, temp_len,
+	jwe.jws.map.len[LJWS_JOSE] = (uint32_t)lws_snprintf(temp, (unsigned int)temp_len,
 			"{\"alg\":\"%s\", \"enc\":\"%s\"}", alg, enc);
-	temp_len -= jwe.jws.map.len[LJWS_JOSE];
+	temp_len -= (int)jwe.jws.map.len[LJWS_JOSE];
 
 	/*
 	 * dup the plaintext into the ciphertext element, it will be
@@ -2065,7 +2065,7 @@ test_akw_encrypt(struct lws_context *context, const char *test_name,
 	n = lws_gencrypto_bits_to_bytes(jwe.jose.enc_alg->keybits_fixed);
 	if (lws_jws_randomize_element(context, &jwe.jws.map, LJWE_EKEY,
 				      lws_concat_temp(temp, temp_len),
-				      &temp_len, n,
+				      &temp_len, (unsigned int)n,
 				      LWS_JWE_LIMIT_KEY_ELEMENT_BYTES)) {
 		lwsl_err("Problem getting random\n");
 		goto bail;
@@ -2078,7 +2078,7 @@ test_akw_encrypt(struct lws_context *context, const char *test_name,
 		goto bail;
 	}
 
-	n = lws_jwe_render_compact(&jwe, compact, compact_len);
+	n = lws_jwe_render_compact(&jwe, compact, (unsigned int)compact_len);
 	if (n < 0) {
 		lwsl_err("%s: lws_jwe_render_compact failed: %d\n",
 			 __func__, n);
