@@ -1,7 +1,7 @@
 /*
  * lws-api-test-lws_tokenize
  *
- * Written in 2010-2019 by Andy Green <andy@warmcat.com>
+ * Written in 2010-2021 by Andy Green <andy@warmcat.com>
  *
  * This file is made available under the Creative Commons CC0 1.0
  * Universal Public Domain Dedication.
@@ -173,6 +173,10 @@ struct expected expected1[] = {
 	expected17[] = {
 		{ LWS_TOKZE_TOKEN, "hello", 5 },
 		{ LWS_TOKZE_ENDED, "", 0 },
+	},
+	expected18[] = {
+		{ LWS_TOKZE_TOKEN, "x=y", 3 },
+		{ LWS_TOKZE_ENDED, "", 0 },
 	}
 ;
 
@@ -204,7 +208,7 @@ struct tests tests[] = {
 	}, {
 		"fr-CH, fr;q=0.9, en;q=0.8, de;q=0.7, *;q=0.5",
 		expected7, LWS_ARRAY_SIZE(expected7),
-		LWS_TOKENIZE_F_RFC7230_DELIMS
+		LWS_TOKENIZE_F_ASTERISK_NONTERM | LWS_TOKENIZE_F_RFC7230_DELIMS
 	},
 	{
 		" Οὐχὶ ταὐτὰ παρίσταταί μοι γιγνώσκειν, ὦ ἄνδρες ᾿Αθηναῖοι, greek",
@@ -253,6 +257,10 @@ struct tests tests[] = {
 	{
 		"# comment1\r\nhello #comment2\r\n#comment3", expected17,
 		LWS_ARRAY_SIZE(expected17), LWS_TOKENIZE_F_HASH_COMMENT
+	},
+	{
+		"x=y", expected18,
+		LWS_ARRAY_SIZE(expected18), LWS_TOKENIZE_F_EQUALS_NONTERM
 	}
 };
 
@@ -494,7 +502,7 @@ int main(int argc, const char **argv)
 		memset(&ts, 0, sizeof(ts));
 		ts.start = tests[n].string;
 		ts.len = strlen(ts.start);
-		ts.flags = tests[n].flags;
+		ts.flags = (uint16_t)tests[n].flags;
 
 		do {
 			e = lws_tokenize(&ts);
@@ -543,7 +551,7 @@ int main(int argc, const char **argv)
 	if (p) {
 		ts.start = p;
 		ts.len = strlen(p);
-		ts.flags = flags;
+		ts.flags = (uint16_t)flags;
 
 		printf("\t{\n\t\t\"%s\",\n"
 		       "\t\texpected%d, LWS_ARRAY_SIZE(expected%d),\n\t\t",
@@ -600,6 +608,131 @@ int main(int argc, const char **argv)
 		printf("\t}\n");
 	}
 
+#if defined(LWS_ROLE_H1) || defined(LWS_ROLE_H2)
+	{
+		time_t t;
+
+		if (lws_http_date_parse_unix("Tue, 15 Nov 1994 08:12:31 GMT", 29, &t)) {
+			lwsl_err("%s: date parse failed\n", __func__);
+			fail++;
+		} else {
+			/* lwsl_notice("%s: %llu\n", __func__, (unsigned long long)t); */
+			if (t != (time_t)784887151) {
+				lwsl_err("%s: date parse wrong\n", __func__);
+				fail++;
+			} else {
+				char s[30];
+
+				if (lws_http_date_render_from_unix(s, sizeof(s), &t)) {
+					lwsl_err("%s: failed date render\n", __func__);
+					fail++;
+				} else {
+					if (!strcmp(s, "Tue, 15 Nov 1994 08:12:31 GMT")) {
+						lwsl_err("%s: date render wrong\n", __func__);
+						fail++;
+					}
+				}
+			}
+		}
+	}
+#endif
+
+	{
+		char buf[24];
+		int m;
+
+		m = lws_humanize(buf, sizeof(buf), 0, humanize_schema_si);
+		if (m != 1 || strcmp(buf, "0")) {
+			lwsl_user("%s: humanize 1 fail '%s' (%d)\n", __func__, buf, m);
+			fail++;
+		}
+		m = lws_humanize(buf, sizeof(buf), 2, humanize_schema_si);
+		if (m != 1 || strcmp(buf, "2")) {
+			lwsl_user("%s: humanize 2 fail '%s' (%d)\n", __func__, buf, m);
+			fail++;
+		}
+		m = lws_humanize(buf, sizeof(buf), 999, humanize_schema_si);
+		if (m != 3 || strcmp(buf, "999")) {
+			lwsl_user("%s: humanize 3 fail '%s' (%d)\n", __func__, buf, m);
+			fail++;
+		}
+		m = lws_humanize(buf, sizeof(buf), 1000, humanize_schema_si);
+		if (m != 4 || strcmp(buf, "1000")) {
+			lwsl_user("%s: humanize 4 fail '%s' (%d)\n", __func__, buf, m);
+			fail++;
+		}
+		m = lws_humanize(buf, sizeof(buf), 1024, humanize_schema_si);
+		if (m != 7 || strcmp(buf, "1.000Ki")) {
+			lwsl_user("%s: humanize 5 fail '%s' (%d)\n", __func__, buf, m);
+			fail++;
+		}
+	}
+
+	if (lws_strcmp_wildcard("allied", 6, "allied")) {
+		lwsl_user("%s: wc 1 fail\n", __func__);
+		fail++;
+	}
+	if (lws_strcmp_wildcard("a*", 2, "allied")) {
+		lwsl_user("%s: wc 2 fail\n", __func__);
+		fail++;
+	}
+	if (lws_strcmp_wildcard("all*", 4, "allied")) {
+		lwsl_user("%s: wc 3 fail\n", __func__);
+		fail++;
+	}
+	if (lws_strcmp_wildcard("all*d", 5, "allied")) {
+		lwsl_user("%s: wc 4 fail\n", __func__);
+		fail++;
+	}
+	if (!lws_strcmp_wildcard("b*", 2, "allied")) {
+		lwsl_user("%s: wc 5 fail\n", __func__);
+		fail++;
+	}
+	if (!lws_strcmp_wildcard("b*ed", 4, "allied")) {
+		lwsl_user("%s: wc 6 fail\n", __func__);
+		fail++;
+	}
+	if (!lws_strcmp_wildcard("allie", 5, "allied")) {
+		lwsl_user("%s: wc 7 fail\n", __func__);
+		fail++;
+	}
+	if (lws_strcmp_wildcard("allie*", 6, "allied")) {
+		lwsl_user("%s: wc 8 fail\n", __func__);
+		fail++;
+	}
+	if (lws_strcmp_wildcard("*llie*", 6, "allied")) {
+		lwsl_user("%s: wc 9 fail\n", __func__);
+		fail++;
+	}
+	if (lws_strcmp_wildcard("*llied", 6, "allied")) {
+		lwsl_user("%s: wc 10 fail\n", __func__);
+		fail++;
+	}
+	if (!lws_strcmp_wildcard("*llie", 5, "allied")) {
+		lwsl_user("%s: wc 11 fail\n", __func__);
+		fail++;
+	}
+	if (!lws_strcmp_wildcard("*nope", 5, "allied")) {
+		lwsl_user("%s: wc 12 fail\n", __func__);
+		fail++;
+	}
+	if (lws_strcmp_wildcard("*li*", 4, "allied")) {
+		lwsl_user("%s: wc 13 fail\n", __func__);
+		fail++;
+	}
+	if (lws_strcmp_wildcard("*", 1, "allied")) {
+		lwsl_user("%s: wc 14 fail\n", __func__);
+		fail++;
+	}
+	if (lws_strcmp_wildcard("*abc*d", 6, "xxabyyabcdd")) {
+		lwsl_user("%s: wc 15 fail\n", __func__);
+		fail++;
+	}
+	if (lws_strcmp_wildcard("ssproxy.n.cn.*", 14, "ssproxy.n.cn.failures")) {
+		lwsl_user("%s: wc 16 fail\n", __func__);
+		fail++;
+	}
+
 	lwsl_user("Completed: PASS: %d, FAIL: %d\n", ok, fail);
 
 	return !(ok && !fail);
